"""This program solves the grade-two non-Newtonian fluid model equations
in a contracting duct with Dirichlet boundary conditions given by
   u=(1-y^2,0) on the left,  u = (0.5(1-0.25*y^2),0) on the right and
u=0 elsewhere.
The boundary conditions for z are posed only on the inflow boundary
This version does Anderson acceleration on u
"""

from mshr import *
from dolfin import *
import math,sys
import write_error
from scipy import linalg as la
from numpy import zeros as zeros
from numpy import float64 as float64
from numpy import int64 as int64
from numpy import ceil  as ceil
from numpy import log10 as log
from numpy.linalg import norm as lanorm
from numpy import inner as npinner
from numpy import max as npmax
from numpy import min as npmin

#import time
# -- -----------------------------------
# -- open an output file
it_out = open('iter_log.txt','w')
# -- -----------------------------------
# -- Set some AA solver parameters
# -- -----------------------------------
betaD     = 1.0  ## 1.0 = no damping
m_max     = 10; ## m_max = 0 means no regularization
m_acc     = 10; ## depth runs from m_acc to m_max (make them same for static m)
start_acc = 1  ## start acc after iteration start_acc
stol      = 1/sqrt(2) ## init. filtering tolerance # 0 is no filtering, 1 forces m=1.
thetaSG   = 1.0 ## only accelerate if theta < thetaSG (thetaSG=1.0: no safeguarding)
# -- -----------------------------------
# -- mesh refinement parameters:
refines = 0 ## uniform refinements
brefs = #9   ## boundary refinements
srefs = #12   ## singularity refinements
# -- -----------------------------------

set_log_active(False)

nu=float(sys.argv[1])
alfaone=float(sys.argv[2])
alfatwo=float(sys.argv[3])
rbufr = 1.0

upright=0.5
right=1.0
lbufr=-1.0
# -- -----------------------------------
you=0.0625 # = 2^{-4} (parameter "U")
# -- -----------------------------------
pdeg=4     # Lagrange poly degree
# -- -----------------------------------

def in_bdry(x):
  return x[0] < lbufr+DOLFIN_EPS

rightprbufr=right+rbufr

# Create mesh and define function space
mesh = Mesh("contracting_duct_M6N4.xml")

# refine mesh uniformly
for j in range(refines):
  mesh = refine(mesh)
  print('h = ',mesh.hmax())

ps1 = Point(0.4,0.5) # near northern boundary 
ps2 = Point(0.4,-0.5) # near southern boundary 
# refine elements near boundary
for j in range(brefs):
  mesh.init(1,2)
  cell_markers = MeshFunction('bool',mesh,mesh.topology().dim())
  cell_markers.set_all(False)
  for c in cells(mesh):
    if c.distance(ps1) < 0.65 or c.distance(ps2)<0.65:
      for f in facets(c):
        if f.exterior():
          cell_markers[c] = True
          break
  mesh = refine(mesh,cell_markers)


# refine elements near singularities
psN = Point(1.0,0.5) # northern singularity
psS = Point(1.0,-0.5)# southern singularity 
pcN = Point(0.0,1.0)
pcS = Point(0.0,-1.0)
for j in range(srefs):
  mesh.init(1,2)
  cell_markers = MeshFunction('bool',mesh,mesh.topology().dim())
  cell_markers.set_all(False)
  for c in cells(mesh):
    if c.contains(psN) or c.contains(psS) or c.contains(pcN) or c.contains(pcS):
      cell_markers[c] = True
  mesh = refine(mesh,cell_markers)


h = mesh.hmax()
hmi = mesh.hmin()
hti = h/(2.0**6)
# -- picture of mesh
#titlestr = 'M6N4_r' + str(refines) +'_br'+ str(brefs) + '_sr'+ str(srefs)+ '.png'
#write_error.save_mesh(mesh,cells,titlestr)

# -- -----------------------------------------------------------------------------
V = VectorFunctionSpace(mesh, "Lagrange", pdeg)
W = FunctionSpace(mesh, "Lagrange", pdeg)
# -- print some initial output to file --------------------------------------------------
istr = 'pdeg = '+ str(pdeg) + \
  ', h: ' + str.format("{0:.4f}",mesh.hmax()) + ', Vdim = ' + str(V.dim()) + ', Wdim ' + str(W.dim()) + ', m_max = ' + str(m_max) + ', stol = ' + str(stol) + '\n\n'
it_out.write(istr)

istr = 'kters\titers\t||du||(H1rel)   ||Mdu||      ||du||' + \
  '\t  theta\t\tmn\t||u||(H1)\n' 
it_out.write(istr)
# -- ------------------------------------------------------------------------------------

# define boundary condition
# at the inlet, L=2, at the outlet L=1
boundaryou  = Expression(("((exp(-kl*(1.0+x[0])*(1.0+x[0]))*fu*(1.0-x[1]*x[1]))+\
                            (exp(-kl*(rtz-x[0])*(rtz-x[0]))*fu*8.0*(0.25-x[1]*x[1])))","0.0"),\
                             rtz=rightprbufr,fu=you,kl=100.0,degree=pdeg)
alfie=4.0*you*you*(3.0*alfaone+2.0*alfatwo)/nu
boundubya = Expression(("0.0","fu*x[1]*exp(-kl*(1.0+x[0])*(1.0+x[0]))"),rtz=rightprbufr,kl=1.0,fu=alfie,degree = pdeg)

w=interpolate(boundubya,V)
atr=you*you*(2.0*alfaone+alfatwo)/nu
aor=2.0*alfaone*you*you/nu
tur=-2.0*you
mcr=you*(-1.0+(you*(3.0*alfaone+alfatwo))/(3.0*nu))

bcyou = DirichletBC(V, boundaryou, "on_boundary")
bcdubya = DirichletBC(V, boundubya, in_bdry)
roipm = 1.0e4

# Solve Stokes
us = TrialFunction(V)
v = TestFunction(V)
# -- -----------------------------------
# -- -----------------------------------
z = Function(V)
ast = inner(grad(us), grad(v))*dx + roipm*div(us)*div(v)*dx
Fs = -inner(div(z),div(v))*dx
us = Function(V)
pdes = LinearVariationalProblem(ast, Fs, us, bcyou)
solvers = LinearVariationalSolver(pdes)
# Scott-Vogelius iterated penalty method
iters = 0; max_iters = 10; div_u_norm = 1
while iters < max_iters and div_u_norm > 1e-10:
    # solve and update z
    solvers.solve()
    z.vector().axpy(roipm, us.vector())
    # find the L^2 norm of div(u) to check stopping condition
    div_u_norm = sqrt(assemble(div(us)*div(us)*dx(mesh)))
    iters += 1
print(" Stokes IPM iter_nos=",iters,"div_u_norm="," %.2e"%div_u_norm)
    
# -- save plot
titlestr = 'u_stokes.png'
write_error.save_soln_pdf(mesh,cells,us[0],titlestr,0.0)

# -- -----------------------------------
## -- initialize X_mat and F_mat-------------------
ndof = V.dim()
X_mat = zeros((ndof,m_max),float64,'F')
F_mat = zeros((ndof,m_max),float64,'F')
colchk= zeros((m_max,),float64,'F') ## for filtering
mn = m_acc
Z_mat = zeros((ndof,m_max),float64,'F')
FZ_mat = zeros((ndof,m_max),float64,'F')
# -- -----------------------------------
# -- u variables (with extras for acc)
# -- -----------------------------------
uu   = TrialFunction(V)
u    = Function(V)
uold = Function(V)
eu   = Function(V)
du   = Function(V)
du1  = Function(V)
# -- -----------------------------------
zold = Function(V)
ez   = Function(V) # -- diff between iters
dz   = Function(V) # -- update
dz1  = Function(V)
# -- -----------------------------------
du1v = as_backend_type(du1.vector()).vec()
euv  = as_backend_type(eu.vector()).vec()
# -- -----------------------------------
wold = Function(V)
## -------------------------------------------------
kters = 0; max_kters = 151; unorm = 1
theta = 1.0; normdu = 1.0; normeu = 1.0 ## initialize for output
incrunorm=1.0
tol = betaD*1e-5
while kters < max_kters and incrunorm > tol and div_u_norm < 1.0:
# -- ---------------------------------------
    z = Function(V)
    a = inner(grad(uu), grad(v))*dx + roipm*div(uu)*div(v)*dx 
    F = inner(w,v)*dx-inner(div(z),div(v))*dx
    # -- -----------------------------------------
    pde = LinearVariationalProblem(a, F, u, bcyou)
    solver = LinearVariationalSolver(pde)
# Scott-Vogelius iterated penalty method
    iters = 0; max_iters = 10; div_u_norm = 1; onorm = 10.0;
    while iters < max_iters and div_u_norm > 1e-10 and div_u_norm/onorm < 0.5 :
    # solve and update z
        solver.solve() # solve for u
        z.vector().axpy(roipm, u.vector())
        # find the L^2 norm of div(u) to check stopping condition
        onorm = div_u_norm
        div_u_norm = sqrt(assemble(div(u)*div(u)*dx(mesh)))
        print ("   IPM iter_no=",iters,"div_u_norm="," %.2e"%div_u_norm)
        iters += 1
    print(" G2 IPM iter_nos=",iters,"div_u_norm="," %.2e"%div_u_norm)
    # -- accelerate for u
    # -- ------------------------------------
    #if m_max > 0 and kters > 0: ## -- defining du1 makes sense 
    if kters > 0: ## -- defining du1 makes sense, do it anyways to record errors
      du1.vector()[:] = u.vector() - uold.vector()
      # -- ------------------------------------
      dz1.vector()[:] = z.vector() - zold.vector()
      # -- ------------------------------------
      normdu = du1.vector().norm('l2') ## coefficient norm
    if m_max > 0 and kters > 1: ## -- update X_mat and F_mat
      # -- dynamic m: do it anyways (unless m_max ==0)
      X_mat[:,1:] = X_mat[:,:-1]
      F_mat[:,1:] = F_mat[:,:-1]
      X_mat[:,0] = euv.copy()
      # -- ---------------------------
      Z_mat[:,1:] = Z_mat[:,:-1]
      FZ_mat[:,1:] = FZ_mat[:,:-1]
      Z_mat[:,0] = ez.vector()
      # -- ---------------------------
      du.vector().axpy(-1.0,du1.vector())
      duv  = as_backend_type(du.vector()).vec()
      # -- ---------------------------
      dz.vector().axpy(-1.0,dz1.vector())
      # -- ---------------------------
      F_mat[:,0] =   duv.copy()
      # -- ---------------------------
      FZ_mat[:,0] =  dz.vector() 
      # -- ---------------------------
      ## ## -----------------------------------------
      #mp = ceil((-log(normdu))) ## use for dynamic depth updates
      #m_acc = int64(min(m_max,max(m_acc, mp) ) )
    mn = min(m_acc, kters-1)
    du.assign(du1)
    dz.assign(dz1)
    if  (kters <= start_acc or mn==0):
      mn = 0
    else:  ## solve l.s. problem and adjust next iterate
      mn = min(m_acc, mn)
      du_range, RR = la.qr_multiply(-F_mat[:,:mn],du1v,\
        overwrite_a=False,overwrite_c=False)
      ## ## -----------------------------------------
      ## now r-filter
      ## ## -----------------------------------------
      if mn > 1:
        for ss in range(mn):
          colchk[ss] = (abs(RR[ss,ss])/lanorm(F_mat[:,ss],2))
        condi = colchk > stol # restrict extrapolation to strongly LI columns
        #print('using columns: ',condi[:])
        if sum(condi) < mn:
          mn =  sum(condi) ## now fix F_mat and X_mat, then recompute QR
          X_mat[:,:mn] = X_mat[:,condi]
          F_mat[:,:mn] = F_mat[:,condi]
          du_range, RR = la.qr_multiply(-F_mat[:,:mn],du1v,\
            overwrite_a=False,overwrite_c=False)
          ## ## -----------------------------------------
          Z_mat[:,:mn] = Z_mat[:,condi]
          FZ_mat[:,:mn] = FZ_mat[:,condi]
          ## ## -----------------------------------------
      ## ## -----------------------------------------
      ## -- now QR is fixed; compute theta
      ## ## -----------------------------------------
      normurange = lanorm(du_range)
      delta_acc   = la.solve_triangular(RR,du_range,trans=0,lower=False, \
        unit_diagonal=False, overwrite_b = True, check_finite=False)
      theta   = sqrt(1.0-(normurange/normdu)**2) 
      ## ## -----------------------------------------
      E_mat  = X_mat[:,:mn] - betaD*F_mat[:,:mn]
      EZ_mat = Z_mat[:,:mn] - betaD*FZ_mat[:,:mn]
      ## ## -----------------------------------------
      #if theta < thetaSG: ## another safeguarding step?
      u.vector()[:] = (uold.vector() + betaD*du1.vector()) - E_mat.dot(delta_acc.T)
      ## ## -----------------------------------------
      z.vector()[:] = (zold.vector() + betaD*dz1.vector()) - EZ_mat.dot(delta_acc.T)
      ## ## -----------------------------------------
    ## ## -----------------------------------------
    eu.vector()[:] = u.vector() - uold.vector();
    ez.vector()[:] = z.vector() - zold.vector();
    norme = lanorm(eu.vector()[:]) 
# -- ---------------------------------------
# -- end accelerate for u
# -- -----------------------------------------------
    # -- updates stol
    # -- -----------------------------------------------
    stol = max(min(1/sqrt(2),sqrt(normdu)),0.1)
    # -- -----------------------------------------------


    kters += 1
    unorm=norm(u,norm_type='H1')
# -- ---------------------------------------------------------------------------------
# -- solve for w 
# -- ---------------------------------------------------------------------------------
    # ultimate Stokes pressure
    piu = project(-div(z), FunctionSpace(mesh, "Lagrange", pdeg - 1))
# -- ---------------------------------------------------------------------------------

    v = TestFunction(V)
    w = TrialFunction(V)

    aw = nu*inner(w,v)*dx + alfaone*inner(grad(w)*u,v)*dx 
# Fw = - div N
    Fw = inner(alfaone*div(grad(u).T*(grad(u)+grad(u).T)),v)*dx \
   +(alfaone+alfatwo)*inner(div((grad(u)+grad(u).T)*(grad(u)+grad(u).T)),v)*dx \
        -inner(grad(u)*u,v)*dx - (alfaone*inner(grad(u).T*grad(piu),v))*dx 

    w = Function(V)
    pdew = LinearVariationalProblem(aw, Fw, w, bcdubya)
    solvew = LinearVariationalSolver(pdew)
    solvew.solve()

# -- ---------------------------------------------------------------------------------
    unormh1=norm(u,norm_type='H1')
    incrunorm=errornorm(uold,u,norm_type='H1',degree_rise=2)/unormh1
    wlinf = norm(w.vector(),'linf')
    print(" nonlin iter_no=",kters," rel change norm="," %.3e"%incrunorm,", mn = ",mn, ", theta = ",theta,"w_inf = ",wlinf)#,"wmax = ",wmax,"wmin = ",wmin)

    uold.assign(u)
    wold.assign(w)
    zold.assign(z)
    # -- print output to file ---------------------------------------------------------
    istr = str.format("{:2}",kters) + '\t' +\
           str.format("{:2}",iters) + '\t' +\
           str.format("{0:.2e}",incrunorm) + '\t' + \
           str.format("{0:6.2e}",normdu) + '\t' + \
           str.format("{0:.2e}",theta) + '\t' + \
           str.format("{0:2}",mn) + '\t' + \
           str.format("{0:.2e}",unormh1) + '\n'
    it_out.write(istr) ## output parameters used to compute output residual
    it_out.flush()  ## to write while code still running

print("m_acc = ",m_acc," m_max = ",m_max," stol = ",stol)       
print("sr","hmax"," U","    nu","   al_one","   al_two"," niter","div_u_norm ")
print(srefs,"%.2f"%h,"%.2e"%you,"%.1e"%nu,"%.1e"%alfaone," ","%.1e"%alfatwo,"  ",kters,"   %.2e"%div_u_norm)
# -- ------------------------------------------------------------------
# -- ------------------------------------------------------------------
# -- save plot
#titlestr = 'u0_r' + str(refines) + '.png'
#write_error.save_soln_pdf(mesh,cells,u[0],titlestr,-0.3)
#titlestr = 'u1_r' + str(refines) + '.png'
#write_error.save_soln_pdf(mesh,cells,u[1],titlestr,-0.3)

# -- compute physical pressure
pr  = project(nu*piu + alfaone*(u[0]*Dx(piu,0) + u[1]*Dx(piu,1)),FunctionSpace(mesh,"Lagrange",pdeg-1))
# -- ------------------------------------------------------------------
# -- define "force boundary" for force calculation --------------------
class force_bdry(SubDomain):
  def inside(self,x,on_boundary):
    return bool(x[0]>(0.0 + DOLFIN_EPS) and (x[0]<1.0 -DOLFIN_EPS) and on_boundary)

class force_bdry_up(SubDomain):
  def inside(self,x,on_boundary):
    return bool(x[0]>(0.0+DOLFIN_EPS) and (x[1]>0.5+DOLFIN_EPS) and on_boundary)

class force_bdry_lo(SubDomain):
  def inside(self,x,on_boundary):
    return bool(x[0]>(0.0+DOLFIN_EPS) and (x[1]<-0.5-DOLFIN_EPS) and on_boundary)


mf0 = MeshFunction("size_t", mesh, mesh.topology().dim() - 1)
f_bdy = force_bdry()
mf0.set_all(0)
f_bdy.mark(mf0, 1)  
ds = Measure('ds',subdomain_data=mf0)

# -- -------------------------------------------------------------------

# -- now do a force computation
n = FacetNormal(mesh)
# -- (1) marks force_bdy
# -- (2) marks upper force_bdy
# -- (3) marks lower force_bdy

#force_int1up = assemble(nu*2.0/sqrt(5.0)*(Dx(u[0],0) + Dx(u[0],1) + Dx(u[1],0))*ds_up(2))
#force_int1lo = assemble(nu*2.0/sqrt(5.0)*(Dx(u[0],0) - Dx(u[0],1) - Dx(u[1],0))*ds_lo(3))


force_int1 = nu*inner(n,(grad(u)+grad(u).T)*Constant((1.0,0.0)))*ds(1) 
force_int23 = pr*n[0]*ds(1)

force = assemble(force_int1) - assemble(force_int23)
print('F/U: ',force/you)

